package com.supervision.service.impl;

import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.lang.Assert;
import cn.hutool.core.util.StrUtil;
import com.supervision.domain.IrKnowledge;
import com.supervision.domain.IrSessionParam;
import com.supervision.domain.IrSqlParam;
import com.supervision.dto.QueryProcessDTO;
import com.supervision.service.IrKnowledgeService;
import com.supervision.service.IrSessionParamService;
import com.supervision.service.IrSqlParamService;
import com.supervision.service.QueryTemplateProcessor;
import com.supervision.util.ParamTypeConverter;
import com.supervision.util.freemark.StringTemplateConfig;
import com.supervision.util.mybatis.RowSqlMapper;
import freemarker.template.TemplateException;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Propagation;
import org.springframework.transaction.annotation.Transactional;

import java.io.IOException;
import java.util.*;
import java.util.function.Supplier;
import java.util.stream.Collectors;

@Slf4j
@Service
@RequiredArgsConstructor
public class QueryTemplateProcessorImpl implements QueryTemplateProcessor {

    private final IrKnowledgeService knowledgeService;

    private final IrSqlParamService sqlParamService;

    private final IrSessionParamService sessionParamService;

    private final RowSqlMapper rowSqlMapper;
    @Override
    //note:重新构建一个事务，与调用方事务隔离
    @Transactional(rollbackFor = Exception.class, propagation = Propagation.REQUIRES_NEW)
    public QueryProcessDTO process(String templateId, String sessionId, Supplier<List<IrSessionParam>> sessionParamSupplier) {

        Assert.notEmpty(templateId, "模板ID不能为空");

        // 根据模板ID查询查询sql模板和结果模板
        IrKnowledge knowledge = knowledgeService.getById(templateId);

        // 校验必要参数
        QueryProcessDTO queryProcessDTO = QueryProcessDTO.builder().state(0).build();
        queryProcessDTO.notNullParam(null, knowledge, "未查询到意图信息");
        queryProcessDTO.notEmptyParam("sqlTemplate", knowledge.getSqlTemplate(), "模板内容不能为空");
        queryProcessDTO.notEmptyParam("resultTemplate", knowledge.getResultTemplate(), "结果模板不能为空");
        queryProcessDTO.notNullParam("resultType", knowledge.getResultType(), "结果类型不能为空");
        if (Integer.valueOf(1).equals(queryProcessDTO.getState())) {
            // 参数校验不通过，返回结果
            return queryProcessDTO;
        }

        // 查询知识查询sql需要的参数
        List<IrSqlParam> paramsList = sqlParamService.lambdaQuery().eq(IrSqlParam::getKnowledgeId, templateId).list();
        // 校验知识查询sql必填参数
        List<IrSessionParam> sessionParams = null == sessionParamSupplier ?
                sessionParamService.lambdaQuery().eq(IrSessionParam::getSessionId, sessionId).list() : sessionParamSupplier.get();
        Set<String> sessionParamsNames = sessionParams.stream().map(IrSessionParam::getParamName).collect(Collectors.toSet());
        paramsList.stream().filter(param -> Integer.valueOf(1).equals(param.getParamRequire())).forEach(param -> {
            queryProcessDTO.notNullParam(param.getParamName(),
                    sessionParamsNames.contains(param.getParamName()) ? true : null, "参数[" + param.getParamName() + "]未设置！");
        });
        if (Integer.valueOf(1).equals(queryProcessDTO.getState())) {
            // 参数校验不通过，返回结果
            return queryProcessDTO;
        }

        // 执行查询sql
        List<Map<String, Object>> maps = null;
        try {
            maps = rowSqlMapper.selectList(knowledge.getSqlTemplate(), getParams(paramsList, sessionParams));
        } catch (Exception e) {
            log.error("执行查询sql失败", e);
            queryProcessDTO.setState(2);
            return queryProcessDTO;
        }

        // 组装返回结果
        wrapperSqlResult(knowledge, maps, queryProcessDTO);

        return queryProcessDTO;
    }

    private void wrapperSqlResult(IrKnowledge knowledge, List<Map<String, Object>> maps, QueryProcessDTO queryProcessDTO) {

        if (Integer.valueOf(1).equals(knowledge.getResultType())) {
            try {
                String process = StringTemplateConfig.getInstance().process(knowledge.getResultTemplate(), CollUtil.getFirst(maps));
                queryProcessDTO.setStringContent(process);
                queryProcessDTO.setContentType(1);
            } catch (IOException | TemplateException e) {
                log.error("执行结果模板失败", e);
                queryProcessDTO.setState(2);
            }
        }
        if (Integer.valueOf(2).equals(knowledge.getResultType())) {
            try {
                String process = StringTemplateConfig.getInstance().process(knowledge.getResultTemplate(), maps);
                queryProcessDTO.setStringContent(process);
                queryProcessDTO.setContentType(1);
            } catch (IOException | TemplateException e) {
                log.error("执行结果模板失败", e);
                queryProcessDTO.setState(2);
            }
        }

        if (Integer.valueOf(3).equals(knowledge.getResultType())) {
            // 如果存在byte类型的数据，数据key统一命名为 byteContent
            queryProcessDTO.setByteContent((byte[]) CollUtil.getFirst(maps).get("byteContent"));
            queryProcessDTO.setContentType(3);
            queryProcessDTO.setByteContentId(CollUtil.getFirst(maps).get("byteContentId").toString());
        }
        // 去除回响应字符串中的特殊字符
        if (StrUtil.isNotEmpty(queryProcessDTO.getStringContent())){
            queryProcessDTO.setStringContent(StrUtil.cleanBlank(queryProcessDTO.getStringContent()));
        }
    }


    private Map<String, Object> getParams(List<IrSqlParam> paramsList, List<IrSessionParam> sessionParams) {
        Map<String, Object> params = new HashMap<>();
        ParamTypeConverter converter = ParamTypeConverter.getInstance();
        for (IrSqlParam param : paramsList) {
            Optional<IrSessionParam> sessionParam = sessionParams.stream().filter(v -> v.getParamName().equals(param.getParamName())).findFirst();
            sessionParam.ifPresent(irSessionParam -> params.put(param.getParamName(), converter.convert(param.getParamType(), irSessionParam.getParamValue())));
        }
        return params;
    }


}
